from parser.Director import Director
from parser.BasicInputBuilder import BasicInputBuilder
from parser.TestingBuilder import TestingBuilder
from parser.Option import Option
from sql.SQLStrategyPostgres import SQLStrategyPostgres
#from sql.PostgresSourceManager import PostgresSourceManager
from sql.HiveSourceManager import HiveSourceManager
from sql.SQLStrategyHive import SQLStrategyHive
#from sql.DBConfiguration import DBConfiguration


class Facade(object):

    def __init__(self):
        self.__director = None
        self.__inputBuilder = None
        self.__outputBuilder = None

    def solve(self, rules, facts, outputPredicates, db, server, usr, pwd):
        self.__director = Director()

        self.__inputBuilder = BasicInputBuilder()
        self.__director.configureInputBuilder(self.__inputBuilder)

        if Option.getInstance().getTesting():
            self.__outputBuilder = TestingBuilder()
            self.__director.configureOutputBuilder(self.__outputBuilder)

        err = self.__director.parseLine(rules)
        err = self.__director.parseLine(facts)

        if not err:
            #if not Option.getInstance().getTesting():
            #    print("*******Tracing*******")
            #    print("Input Program: \n" + self.__inputBuilder.getProgram().__str__())
            program = self.__inputBuilder.getProgram()

            if not program.isManagedProgram():
                return "The system can't execute this program."

            '''HIVE'''
            srcManager = HiveSourceManager(db, server, usr, pwd)
            sqlStrategy = SQLStrategyHive(program, srcManager)

            '''POSTGRES'''
            #srcManager = PostgresSourceManager(db, server, usr, pwd)
            #sqlStrategy = SQLStrategyPostgres(program, srcManager)

            #DBConfiguration.getInstance().setDBConfig(sqlStrategy)

            for i in range(program.getNumberOfComponents()):
                #if not Option.getInstance().getTesting():
                #    print("component {0}:".format(i))
                subProgram = program.getComponentSubProgram(i)
                if subProgram is not None:
                    # exit subprogram
                    for rule in subProgram.getExitRules():
                        #if not Option.getInstance().getTesting():
                        #    print("Exit rule: {0}".format(rule.__str__()))
                        sqlStatement = sqlStrategy.rewriteExitRule(rule)
                        #if not Option.getInstance().getTesting():
                        #    print("{0}".format(sqlStatement))
                        srcManager.executeStatement(sqlStatement)
                    recursionLevel = 0
                    #if not Option.getInstance().getTesting():
                    #    print("\nStart recursive subProgram.")
                    finished = True
                    rowsAffected = []
                    while True:
                        rules = subProgram.getRecursiveRules()
                        if not rules:
                            #if not Option.getInstance().getTesting():
                            #    print("There aren't recursive rules.")
                            break
                        for r in rowsAffected:
                            if r != 0 and r is not None:
                                rowsAffected.clear()
                                finished = False
                        if finished and len(rowsAffected) != 0:
                            #if not Option.getInstance().getTesting():
                            #    print("There aren't new inserted tuples.")
                            break
                        finished = True
                        ruleIdx = 0
                        for rule in subProgram.getRecursiveRules():
                            #if not Option.getInstance().getTesting():
                            #    print("Recursive rule: {0}".format(rule.__str__()))
                            sqlStatement = sqlStrategy.rewriteRecursiveRule(ruleIdx, rule, recursionLevel, program)
                            #if not Option.getInstance().getTesting():
                            #    print("{0}".format(sqlStatement))
                            cursor = srcManager.executeStatement(sqlStatement)

                            #if isinstance(DBConfiguration.getInstance().getDBConfig(), SQLStrategyPostgres):
                            #    rowsAffected.append(sqlStrategy.getNumberRowsInserted(cursor))

                            #el
                            #if isinstance(DBConfiguration.getInstance().getDBConfig(), SQLStrategyHive):
                            sqlStatement = sqlStrategy.getNumberRowsInserted(rule.getHead(), recursionLevel+1)
                            result = srcManager.executeStatement(sqlStatement)
                            rowsAffected.append(result.fetchone()[0])

                            cursor.close()
                            ruleIdx = ruleIdx + 1
                        recursionLevel = recursionLevel + 1
            srcManager.commit()
            results = []

            if outputPredicates != "":
                name = outputPredicates[:outputPredicates.find("/")]
                arity = outputPredicates[outputPredicates.find("/") + 1:]
                predicate = None
                for rule in program._getRules():
                    pName = rule.getHead()[0].getPredicate().getName()
                    pArity = rule.getHead()[0].getPredicate().getArity()
                    if pName == name and str(pArity) == arity:
                        predicate = rule.getHead()[0].getPredicate()
                        break
                if predicate is not None:
                    sqlStatement = sqlStrategy.getTuples(predicate)
                    cursor = srcManager.executeStatement(sqlStatement)
                    for r in cursor:
                        results.append(r)
                    cursor.close()
                else:
                    return "Not valid outputPredicate."

            elif Option.getInstance().getTesting() and outputPredicates=="":
                targetsName = []
                for rule in program._getRules():
                    for predicate in rule.getPredicates():
                        name = predicate.getName()
                        if name not in targetsName:
                            targetsName.append(name)
                            sqlStatement = sqlStrategy.getTuples(predicate)
                            cursor = srcManager.executeStatement(sqlStatement)
                            self.__outputBuilder.onPredicate(name, cursor)
                            sqlStatement = sqlStrategy.dropTable(predicate)
                            cursor = srcManager.executeStatement(sqlStatement)
                            cursor.close()
                srcManager.commit()
                results = self.__outputBuilder.getResults()
                results.sort()

        #if not Option.getInstance().getTesting():
        #    print("Depgraph:")
        #    print(program.getDepGraph().__str__())
        #    print("Output Predicates: " + outputPredicates)
        #    print("DBName: " + db)
        #    print("Host: " + server)
        #    print("Username: " + usr)
        #    print("Password: " + pwd)

        return results
