'''
Created on Apr 10, 2018

@author: jessica
'''
from depgraph.Depgraph import Depgraph
from datalog.Literal import NafLiteral
from datalog.Rule import Rule
from datalog.SubProgram import SubProgram

class Program(object):

    def __init__(self):
        self.__rules = []
        self.__depgraph = Depgraph()
        self.__compSubPrograms = {}
        self.__managedProgram = True

    def isManagedProgram(self):
        return self.__managedProgram

    def setManagedProgram(self, managed):
        self.__managedProgram=managed

    def _getRules(self):
        return self.__rules

    def addRule(self, rule):
        self.__rules.append(rule)
        for at in rule.getHead():
            trg = at.getPredicate().getIndex()
            if len(rule.getBody()) == 0:
                self.__depgraph.addVertex(trg)
            # add dependencies
            for lit in rule.getBody():
                if isinstance(lit,NafLiteral):
                    src = lit.getPredicate().getIndex()
                    neg = lit.getNegatedAsFailure()
                    self.__depgraph.addDep(src,trg,neg)
        
    def addRules(self, rules):
        for rule in rules:
            self.addRule(rule)

    def isStratified(self):
        return self.__depgraph.isStratified()
    
    def getNumberOfComponents(self):
        return len(self.__depgraph.getSCCList())
    
    def getComponentSubProgram(self,compIdx):
        if len(self.__compSubPrograms) == 0:
            self.__computeComponentSubPrograms()
        if compIdx in self.__compSubPrograms: 
            return self.__compSubPrograms[compIdx]
        else:
            return None

    def getDepGraph(self):
        return self.__depgraph
    
    def getComponent(self, pred):
        return self.__depgraph.getVertexComponent(pred.getIndex())

    def __computeComponentSubPrograms(self):
        for rule in self.__rules:
            #no disjunction at the moment
            assert len(rule.getHead()) == 1
            headAtom = rule.getHead()[0]
            headComp = self.__depgraph.getVertexComponent(headAtom.getPredicate().getIndex())
            isRecursive = False
            for lit in rule.getBody():
                if isinstance(lit, NafLiteral):
                    if headComp == self.__depgraph.getVertexComponent(lit.getPredicate().getIndex()):
                        isRecursive = True
            self.__addToComponentSubProgram(headComp,rule,isRecursive)

    def __addToComponentSubProgram(self,comp,rule,isRecursive):
        if comp not in self.__compSubPrograms:
            s = SubProgram()
            s.addRule(rule,isRecursive)
            self.__compSubPrograms[comp] = s
        else:
            self.__compSubPrograms[comp].addRule(rule,isRecursive)
    
    def __str__(self):
        s = ""
        for rule in self.__rules:
            s += rule.__str__() + "\n"
        return s
