'''
Created on Apr 10, 2018

@author: jessica
'''

class Term(object):

    def __init__(self, name = None):
        if name == None:
            name = ""
        else:
            self.__name = name

    def setName(self, value):
        self.__name = value

    def getName(self):
        return self.__name

    def getVariables(self):
        return set()

    def __str__(self):
        return self.__name
    
    def __hash__(self):
        return hash(self.getName())
    
class Variable(Term):
    
    def __init__(self, name = None):
        super(Variable, self).__init__(name)
        
    def getVariables(self):
        variables = set()
        variables.add(self)
        return variables
    
    def __eq__(self, other):
        return isinstance(other,Variable) and self.getName() == other.getName()
    
    def __hash__(self):
        return hash(self.getName())

class AnonymousVariable(Variable):
    
    def __init__(self, name = None):
        super(AnonymousVariable, self).__init__(name)
    
    def __eq__(self, other):
        return isinstance(other,AnonymousVariable) and self.getName() == other.getName()
    
    def __hash__(self):
        return hash(self.getName())
        
class Constant(Term):
    
    def __init__(self, name = None):
        super(Constant, self).__init__(name)
    
    def __eq__(self, other):
        return isinstance(other,Constant) and self.getName() == other.getName()
    
    def __hash__(self):
        return hash(self.getName())
        
class SymbolicConstant(Constant):
    
    def __init__(self, name = None):
        super(SymbolicConstant, self).__init__(name)
    
    def __eq__(self, other):
        return isinstance(other,SymbolicConstant) and self.getName() == other.getName()
    
    def __hash__(self):
        return hash(self.getName())
        
class StringConstant(Constant):
    
    def __init__(self, name = None):
        super(StringConstant, self).__init__(name)
    
    def __eq__(self, other):
        return isinstance(other,StringConstant) and self.getName() == other.getName()
    
    def __hash__(self):
        return hash(self.getName())    
        
class IntegerConstant(Constant):
    
    def __init__(self, name = None, negative = False):
        super(IntegerConstant, self).__init__(name)
        self.__negative = negative

    def setNegative(self, value):
        self.__negative = value

    def isNegative(self):
        return self.__negative
    
    def __eq__(self, other):
        return isinstance(other,IntegerConstant) and self.getName() == other.getName() and self.isNegative() == other.isNegative()
    
    def __hash__(self):
        if self.__negative:
            return -hash(self.getName())
        else:
            return hash(self.getName()) 

class FunctionalTerm(Term):
 
    def __init__(self, functor = None, terms = None):
        super(FunctionalTerm, self).__init__(functor)
        if terms == None:
            self.__terms = []
        else:
            self.__terms = terms
        
    def __str__(self):
        s = self.getName()
        if len(self.__terms):
            s += "("
            for term in self.__terms:
                s += term.__str__() + ","
            s = s[:-1]
            s += ")"
        return s
    
    def __eq__(self, other):
        return isinstance(other,FunctionalTerm) and self.getName() == other.getName() and self.__terms == other.__terms
    
    def __hash__(self):
        h = hash(self.getName())
        for term in self.__terms:
            h = h*hash(term)
        return h
    
    def getVariables(self):
        variables = set()
        for term in self.__terms:
            v = term.getVariables()
            for var in v:
                variables.add(var)
        return variables
    