from tarjan import tarjan
from datalog.Predicate import Predicate

class Depgraph(object):
    
    def __init__(self):
        self.__vertexNumber = 0
        self.__graph = {}
        self.__label2vertex = {}
        self.__edge2label = {}
        self.__sccList = []
        self.__vertex2component = {}
        self.__compHasNegation = []
    
    def addDep(self, src, trg, neg = False):
        vSrc = self.__getVertex(src)
        vTrg = self.__getVertex(trg)
        if vSrc not in self.__graph:
            self.__graph[vSrc] = [vTrg]
            self.__edge2label[(vSrc,vTrg)] = neg
        elif vTrg not in self.__graph[vSrc]:
            self.__graph[vSrc].append(vTrg)
            self.__edge2label[(vSrc,vTrg)] = neg
        elif neg == True:
            self.__edge2label[(vSrc,vTrg)] = neg
    
    def addVertex(self, v):
        vertex = self.__getVertex(v)
        if vertex not in self.__graph:
            self.__graph[vertex] = []

    def isEdge(self, src, trg):
        vSrc = self.__getVertex(src)
        vTrg = self.__getVertex(trg)
        if vSrc in self.__graph and vTrg in self.__graph[vSrc]:
            return True
        return False

    def isNegEdge(self, src, trg):
        vSrc = self.__getVertex(src)
        vTrg = self.__getVertex(trg)
        if vSrc in self.__graph and vTrg in self.__graph[vSrc]:
            assert (vSrc,vTrg) in self.__edge2label
            if self.__edge2label[(vSrc,vTrg)] == True:
                return True
        return False
        
    def numVertices(self):
        return self.__vertexNumber
    
    def getSCCList(self):
        if len(self.__sccList) == 0:
            self.__computeSCCs()
        return self.__sccList

    def getVertexComponent(self,vertex):
        if len(self.__sccList) == 0:
            self.__computeSCCs()
        v = self.__getVertex(vertex)
        assert v in self.__vertex2component
        return self.__vertex2component[v]            

    def isStratified(self):
        if len(self.__sccList) == 0:
            self.__computeSCCs()
        for idx in range(len(self.__sccList)):
            if self.isStratified(idx) == False:
                return False
        return True
    
    def isStratified(self,compIdx):
        if len(self.__sccList) == 0:
            self.__computeSCCs()
        if len(self.__compHasNegation) == 0:
            self.__computeNegativeEdges()
        assert compIdx >= 0 and compIdx < len(self.__compHasNegation)
        return not(self.__compHasNegation[compIdx])
    
    def __computeSCCs(self):
        assert len(self.__sccList) == 0
        self.__sccList = tarjan(self.__graph) 
        #components are returned in a reverse topological order
        self.__sccList = self.__sccList[::-1]
        nSCCs = len(self.__sccList)
        for x in range(nSCCs):
            for y in self.__sccList[x]:
                assert y >= 0 and y < self.__vertexNumber
                self.__vertex2component[y] = x        

    def __getVertex(self,label):
        if label in self.__label2vertex:
            return self.__label2vertex[label]
        else:
            self.__label2vertex[label] = self.__vertexNumber
            self.__vertexNumber += 1
            return self.__label2vertex[label]

    def __computeNegativeEdges(self):
        if len(self.__sccList) == 0:
            self.__computeSCCs()
        for i in range(len(self.__sccList)):
            self.__compHasNegation.append(False)
        for edge in self.__edge2label:
            if self.__edge2label[edge] == True:
                src = edge[0]
                trg = edge[1]
                if self.__vertex2component[src] == self.__vertex2component[trg]:
                    compIdx = self.__vertex2component[src]
                    self.__compHasNegation[compIdx] = True
    
    def __str__(self):
        s = ""
        for node in self.__label2vertex:
            vertex = self.__label2vertex[node]
            s += "Pred " + Predicate.getNameFromIdx(node) + "/"+ str(Predicate.getArityFromIdx(node))
            s += "(vertex: " + str(vertex) + ", comp: " + str(self.__vertex2component[vertex]) + "): ["

            if vertex in self.__graph:
                for dep in self.__graph[vertex]:
                    s += dep.__str__() +","
                if len(self.__graph[vertex]) > 0:
                    s = s[:-1]

            s += "]\n"
        return s