'''
Created on Apr 9, 2018

@author: jessica
'''
from datalog.Term import Variable, StringConstant, SymbolicConstant,\
    IntegerConstant, AnonymousVariable, FunctionalTerm
from datalog.Predicate import Predicate
from datalog.Atom import ClassicalAtom, BuiltInAtom
from datalog.Rule import Rule
from datalog.Program import Program
from parser.InputBuilder import InputBuilder
from datalog.Literal import NafLiteral
from parser.ErrorMessage import ErrorMessage

class BasicInputBuilder(InputBuilder):
    
    def __init__(self):
        self.__terms = []
        self.__body = []
        self.__head = []
        self.__rules = []
        self.__program = Program()

    def getProgram(self):
        return self.__program
    
    def onProgram(self):
        self.__program.addRules(self.__rules)
        self.__rules = []
    
    def onRule(self):
        rule = Rule(self.__head, self.__body)
        if not rule.isSafe():
            ErrorMessage.onError("Safety Error in rule: " + rule.__str__())
        self.__rules.append(rule)
        self.__head = []
        self.__body = []
    
    def onBodyLiteral(self):
        if self.__nafLiteral != None:
            self.__body.append(self.__nafLiteral)
            self.__nafLiteral = None
        if self.__atom != None and isinstance(self.__atom, BuiltInAtom):
            self.__body.append(self.__atom)
            self.__atom = None
    
    def onHeadAtom(self):
        self.__head.append(self.__atom)
        
    def onAtom(self):
        pass
    
    def onBuiltInAtom(self):
        self.__program.setManagedProgram(False)
        self.__atom = BuiltInAtom(self.__terms, self.__binop)
        self.__terms = []
        self.__binop = None
    
    def onBinaryOperation(self, binop):
        self.__binop = binop
        
    def onClassicalAtom(self, stronglyNegated):
        if stronglyNegated:
            self.__predicateName = "-"+self.__predicateName
        predicate = Predicate(self.__predicateName, len(self.__terms))
        self.__atom = ClassicalAtom(predicate, self.__terms, stronglyNegated)
        self.__terms = []
        self.__predicateName = ""
        
    def onNafLiteral(self, negatedAsFailure):
        if negatedAsFailure==True:
            self.__program.setManagedProgram(False)
        if isinstance(self.__atom, ClassicalAtom):
            self.__nafLiteral = NafLiteral(self.__atom.getPredicate(), self.__atom.getTerms(), self.__atom.isStronglyNegated(), negatedAsFailure)
            self.__atom = None
        
    def onPredicateName(self, value):
        self.__predicateName = value
    
    def onTerm(self, value):
        self.__terms.append(self.__newTerm(value))

    def onUnknownVariable(self, value):
        self.__terms.append(AnonymousVariable(value))
        
    def onFunctionalTerm(self, value, arity):
        self.__program.setManagedProgram(False)
        terms = self.__terms[-int(arity):]
        functionalTerm = FunctionalTerm(value, terms)
        self.__terms = self.__terms[:-int(arity)]
        self.__terms.append(functionalTerm)

    def __newTerm(self, value):
        term = None
        if value[0] >= 'A' and value[0] <='Z' : # Variable
            term = Variable(value)
        elif value[0] == '\"' and value[-1] == '\"':
            term = StringConstant(value)
        elif value[0] >= 'a' and value[0] <='z':   # String constant
            term = SymbolicConstant(value)
        else:
            term = IntegerConstant(value, int(value)<0)
        return term

    def illegalToken(self):
        self.__program.setManagedProgram(False)
