'''
Created on Apr 9, 2018

@author: jessica
'''
import ply.yacc as yacc

from parser.lex import tokens
from parser.ErrorMessage import ErrorMessage

start = 'program'

global director

def p_program(p):
    '''program : rules'''
    p[0] = p[1]
    director.getInputBuilder().onProgram()
    
def p_rules_list(p):
    '''rules : rules rule_types'''
    p[0] = p[1] + [p[2]]

def p_rules(p):
    '''rules :  normal_rule 
            | constraint
            | fact'''
    p[0] = [p[1]]
    
def p_rule_types(p):
    '''rule_types :  normal_rule 
            | constraint
            | fact'''
    p[0] = [p[1]]
    
def p_rule(p):
    '''normal_rule :  head CONS body DOT'''
    p[0] = p[1] + p[3]
    director.getInputBuilder().onRule()

def p_fact(p):
    '''fact : head DOT'''
    p[0] = p[1]
    director.getInputBuilder().onRule()

def p_constraint(p):
    '''constraint :  CONS body DOT'''
    p[0] = p[2]
    director.getInputBuilder().illegalToken()
    director.getInputBuilder().onRule()
    
def p_head(p): 
    '''head : classic_atom'''
    p[0] = [p[1]]
    director.getInputBuilder().onHeadAtom()
    
def p_body(p):
    '''body : conjunction'''
    p[0] = [p[1]]
    
def p_conjunction_list(p):
    '''conjunction     : conjunction COMMA naf_literal''' 
    p[0] = p[1] + [p[3]]
    director.getInputBuilder().onBodyLiteral()

def p_conjunction(p):
    '''conjunction     : naf_literal'''
    p[0] = [p[1]]
    director.getInputBuilder().onBodyLiteral()
    
def p_naf_literal_naf(p):
    '''naf_literal : NAF classic_atom'''
    p[0] = p[2]
    director.getInputBuilder().onNafLiteral(True)

def p_naf_literal(p):
    '''naf_literal     : classic_atom'''
    p[0] = p[1]
    director.getInputBuilder().onNafLiteral(False)
    
def p_naf_literal_built_in(p):
    '''naf_literal     : builtin_atom'''
    p[0] = p[1]
    director.getInputBuilder().onBuiltInAtom()

def p_builtin_atom(p):
    '''builtin_atom     : term binop term'''
    p[0] = p[1] + p[2] + p[3]

def p_binop(p):
    '''binop    : EQUAL
                | UNEQUAL
                | LESS
                | GREATER
                | LESS_OR_EQ
                | GREATER_OR_EQ'''
    p[0] = p[1]
    director.getInputBuilder().onBinaryOperation(p[0])
    
def p_classic_atom_neg(p):
    '''classic_atom  : NEG atom'''
    p[0] = p[2] 
    director.getInputBuilder().onClassicalAtom(True)

def p_classic_atom(p):    
    '''classic_atom  : atom'''
    p[0] = p[1]
    director.getInputBuilder().onClassicalAtom(False)
    
def p_atom(p):
    '''atom : IDENTIFIER PARAM_OPEN terms PARAM_CLOSE '''
    p[0] = [p[1]] + p[3]
    director.getInputBuilder().onPredicateName(p[1])
    director.getInputBuilder().onAtom()

def p_terms_term(p):
    '''terms : term'''
    p[0] = [p[1]]
    
def p_terms_terms(p):
    '''terms : term COMMA terms'''
    p[0] = [p[1]] + p[3] 
    
def p_term(p):
    '''term : basic_term '''
    p[0] = p[1]
    
def p_basic_term(p):
    '''basic_term : ground_term 
                | variable_term
                | functional_term'''
    p[0] = p[1]

def p_ground_term(p):
    '''ground_term : IDENTIFIER
                | STRING
                | NUMBER 
                | NEG NUMBER'''
    p[0] = p[1]
    director.getInputBuilder().onTerm(p[0])

def p_variable_term(p):
    '''variable_term : basic_variable_term 
                | anonymous_term'''
    p[0] = p[1]
    
def p_basic_variable_term(p):
    '''basic_variable_term : VARIABLE'''
    p[0] = p[1]
    director.getInputBuilder().onTerm(p[0])

def p_anonymous_term(p):
    '''anonymous_term : ANON_VAR'''
    p[0] = p[1]
    director.getInputBuilder().onUnknownVariable(p[0])

def p_functional_term(p):
    '''functional_term : IDENTIFIER PARAM_OPEN terms PARAM_CLOSE'''
    p[0] = [p[1]] + p[3]
    director.getInputBuilder().onFunctionalTerm(p[1],len(p[3]))

def p_error(p):
    if p != None:
        director.getInputBuilder().illegalToken()
        #ErrorMessage.onError("Aborting due to parser errors. Illegal token %s." % p.value)

def yyparse(d, line):
    global director
    director = d
    parser = yacc.yacc(start='program')
    parser.parse(line)
    
