from sql.SourceManager import SourceManager
from parser.Option import Option
from pyhive import hive


class HiveSourceManager(SourceManager):

    def __init__(self, db, server, usr, pwd):
        super(HiveSourceManager, self).__init__()
        self.__connection = None
        try:
            ip = server[:server.find(":")]
            port = server[server.find(":")+1:]
            self.__connection = hive.Connection(host=ip, username=usr, password=pwd, port=port, database=db, auth='CUSTOM')
        except Exception as e:
            print(e)

    def executeStatement(self, statement, params=None):
        cur = self.__connection.cursor()
        try:
            if params:
                cur.execute(statement, params)
            else:
                cur.execute(statement)
            return cur
        except Exception as e:
            print(e)

    def getTableSchema(self, pred):  # if pred does not match any table, create a matching table
        name = pred.getName()
        arity = pred.getArity()
        with self.__connection.cursor() as cur:
            try:
                sql = "DESCRIBE " + name.__str__()
                cur.execute(sql)
                rs1 = cur.fetchall()
                nColumns = len(rs1)
                if len([x for x in rs1 if x[0] == self.getTimeStampColumnName()]) > 0:  # this column is added by our system in order to handle recursive evaluation
                    nColumns = nColumns - 1
                else:
                    s = """ALTER TABLE {} ADD COLUMN {} timestamp """.format(name.__str__(), self.getTimeStampColumnName())
                    cur.execute(s)
                if len([x for x in rs1 if x[0] == self.getRecursionLevelColumnName()]) > 0:  # this column is added by our system in order to handle recursive evaluation
                    nColumns = nColumns - 1
                else:
                    s = """ALTER TABLE {} ADD COLUMN {} varchar (50)""".format(name.__str__(), self.getRecursionLevelColumnName())
                    cur.execute(s)
                if nColumns == arity:
                    return [t[0] for t in rs1]  # return the existent schema'''
            except:
                # create the table
                '''CREATE TABLE c( 
                    attr0 VARCHAR(100),
                    attr1 VARCHAR(100),
                    executiontime TIMESTAMP,
                    reclevel VARCHAR(100),
                    primary key(attr0, attr1, executiontime) disable novalidate
                );'''
                pk = "PRIMARY KEY ("
                sql = "CREATE TABLE " + name + " ("
                for i in range(arity):
                    attrName = "attr" + str(i)
                    sql += attrName + " VARCHAR(50),"  # only strings are supported
                    pk += attrName + ","
                sql += self.getTimeStampColumnName() + " TIMESTAMP,"
                pk += self.getTimeStampColumnName()
                pk += ")"
                sql += self.getRecursionLevelColumnName() + " VARCHAR(50),"
                sql = sql[:-1]
                sql += ", " + pk + " disable novalidate)"
                sql += "STORED AS ORC"
                if not Option.getInstance().getTesting():
                    print(sql)
                try:
                    cur.execute(sql)
                except Exception as e:
                    print(e)
                return ["attr{}".format(i) for i in range(arity)]  # return the created schema

    def commit(self):
        # COMMIT are not yet supported by Hive. All language operations are auto-commit.
        pass
