from sql.SourceManager import SourceManager
from sql.DBConfiguration import DBConfiguration
import psycopg2


class PostgresSourceManager(SourceManager):

    def __init__(self, db, server, usr, pwd):
        super(PostgresSourceManager, self).__init__()
        self.__connection = None
        try:
            self.__connection = psycopg2.connect(dbname=db, host=server, user=usr, password=pwd)
        except Exception as e:
            print(e)

    def executeStatement(self, statement, params=None):
        cur = self.__connection.cursor()
        try:
            if params:
                cur.execute(statement, params)
            else:
                cur.execute(statement)
            return cur
        except Exception as e:
            print(e)

    def commit(self):
        self.__connection.commit()

    def getTableSchema(self, pred):
        name = pred.getName()
        arity = pred.getArity()
        with self.__connection.cursor() as cur:
            cur.execute("""
                SELECT count(*)
                FROM information_schema.tables
                WHERE table_schema='public'
                    AND table_type='BASE TABLE'
                    AND table_name=%s""",
                        (name.__str__(),))
            rs1 = cur.fetchone()
            if rs1[0] == 1:
                cur.execute("""
                    SELECT column_name, data_type, character_maximum_length
                    FROM information_schema.columns
                    WHERE table_name=%s""",
                            (name.__str__(),))
                rs2 = cur.fetchall()
                nColumns = len(rs2)
                if (len([x for x in rs2 if x[
                                               0] == self.getTimeStampColumnName()]) > 0):  # this column is added by our system in order to handle recursive evaluation
                    nColumns = nColumns - 1
                else:
                    s = """
                        ALTER TABLE {}
                        ADD COLUMN {} varchar DEFAULT null""".format(
                        name.__str__(), self.getRecursionLevelColumnName())
                    cur.execute(s)
                if (len([x for x in rs2 if x[
                                               0] == self.getRecursionLevelColumnName()]) > 0):  # this column is added by our system in order to handle recursive evaluation
                    nColumns = nColumns - 1
                else:
                    s = """
                        ALTER TABLE {}
                        ADD COLUMN {} varchar DEFAULT null""".format(
                        name.__str__(), self.getRecursionLevelColumnName())
                    print(s)
                    cur.execute(s)
                    # cur.execute("""
                    #    ALTER TABLE %s
                    #    ADD COLUMN %s varchar NOT NULL DEFAULT '0'""",
                    #    (name.__str__(),self.getRecursionLevelColumnName(),))
                if nColumns == arity:
                    return [t[0] for t in rs2]  # return the existent schema
            # create the table
            pk = "PRIMARY KEY ("
            sql = "CREATE TABLE " + name + " ("
            for i in range(arity):
                attrName = "attr" + str(i)
                sql += attrName + " varchar,"  # only strings are supported
                pk += attrName + ","
            sql += self.getTimeStampColumnName() + " timestamp without time zone DEFAULT '0001-01-01 00:00:00',"
            pk += self.getTimeStampColumnName()
            pk += ")"
            sql += self.getRecursionLevelColumnName() + " varchar DEFAULT null,"
            sql = sql[:-1]
            sql += ", " + pk + ");"
            try:
                cur.execute(sql)
            except Exception as e:
                print(e)
            return ["attr{}".format(i) for i in range(arity)]  # return the created schema
