from sql.SQLStrategy import SQLStrategy
from datalog.Term import Variable, Constant, FunctionalTerm
from datetime import datetime
import math

class SQLStrategyPostgres(SQLStrategy):
    
    def __init__(self, program, srcManager):
        super(SQLStrategyPostgres,self).__init__(program,srcManager)
        self._timestamp = "'"+datetime.now().strftime("%d/%m/%Y %H:%M:%S")+"'::timestamp without time zone"
        self._originTimeStamp = "'01/01/0001 00:00:00'::timestamp without time zone"

    def _getTimeStamp(self):
        return self._timestamp

    def getTuples(self,predicate):
        #SELECT *
        #FROM public.a
        #WHERE a.executiontime = '2020-01-15 20:28:58'
        trgTable = self._getTableName(predicate)
        sql = "SELECT * "
        sql += "FROM "+trgTable
        sql += " WHERE "+trgTable+"."+self._srcManager.getTimeStampColumnName()+" = "+self._getTimeStamp()
        return sql

    def dropTable(self, predicate):
        trgTable = self._getTableName(predicate)
        sql = "DROP TABLE "+trgTable
        return sql

    def _getExitRuleInSQL(self):
        '''INSERT INTO location
            (SELECT DISTINCT t0.attr0, '22/01/2020 12:16:30'::timestamp without time zone AS const0, '0' AS const1
             FROM block AS t0
             WHERE (t0.executiontime = '22/01/2020 12:16:30'::timestamp without time zone OR
                t0.executiontime = '01/01/0001 00:00:00'::timestamp without time zone) AND NOT EXISTS
                (SELECT *
                 FROM location
                 WHERE(location.executiontime = '22/01/2020 12:16:30'::timestamp without time zone OR
                    location.executiontime = '01/01/0001 00:00:00'::timestamp without time zone) AND t0.attr0 = location.attr0))'''
        trgTable = self._getTableName(self._targetPredicate)
        constCounter = 0
        sqlInsert = "INSERT INTO " + trgTable
        sqlSelect = "SELECT DISTINCT "
        notExistsClause = ""
        columnClause = ""
        for i in range(len(self._termsToSelect)):
            t = self._termsToSelect[i]
            if isinstance(t,Variable):
                assert t in self._varsFrom
                coords = self._varsFrom[t]
                srcIndex = coords[0]
                colIndex = coords[1]
                alias = self._getTableAlias(srcIndex)
                colName = self._getAttrName(srcIndex,colIndex)
                sqlSelect += alias+"."+colName+","
                columnClause += alias + "." + colName + "=" + trgTable + "." + self._targetColumns[i] + " AND "
            else:
                if isinstance(t,FunctionalTerm):
                    print("WARNING: functional terms are handled as standard strings.")
                sqlSelect += "'"+t.__str__().replace("\"","")+"' AS const"+str(constCounter)+"," #only strings are supported
                notExistsClause += trgTable+"."+self._targetColumns[i]+" = '"+t.__str__().replace("\"","")+"' AND "
                constCounter += 1

        # OK
        notExistsClause += "("+trgTable+"."+self._srcManager.getTimeStampColumnName()+" = "+self._getTimeStamp()+" OR ";
        notExistsClause += trgTable+"."+self._srcManager.getTimeStampColumnName()+" = "+self._originTimeStamp+"))";
        sqlSelect += self._getTimeStamp()+" AS const"+str(constCounter)+","
        constCounter += 1
        sqlSelect += "'0' AS const"+str(constCounter)+"," #exit rules can infer level_0-tuples only

        sqlSelect = sqlSelect[:-1]
        sqlFrom = ""
        sqlWhere = ""
        if len(self._bodyPreds) == 0:
            # se è un fatto entra qui!
            sqlWhere = " WHERE NOT EXISTS ( SELECT * FROM "+ trgTable + " WHERE "+ notExistsClause;

        if len(self._bodyPreds) > 0:
            sqlFrom += " FROM "
            clause = ""
            for i in range(len(self._bodyPreds)):
                srcTable = self._getTableName(self._bodyPreds[i])
                alias = self._getTableAlias(i)
                sqlFrom += srcTable + " AS " + alias +","
                clause += "("+alias + "." + self._srcManager.getTimeStampColumnName() + " = " + self._getTimeStamp() + " OR ";
                clause += alias + "." + self._srcManager.getTimeStampColumnName() + " = "+self._originTimeStamp+") AND ";

            sqlFrom = sqlFrom[:-1]
            sqlWhere += " WHERE " + clause
            cond = 0
            for t in self._bodyTermPos:
                if t.getName() != "_":
                    if isinstance(t,Variable):
                        coords = self._bodyTermPos[t]
                        if len(coords) > 1:
                            srcIndexBase = coords[0][0]
                            colIndexBase = coords[0][1]
                            aliasBase = self._getTableAlias(srcIndexBase)
                            colNameBase = self._getAttrName(srcIndexBase,colIndexBase)
                            for i in range(len(coords[1:])):
                                srcIndex = coords[i+1][0]
                                colIndex = coords[i+1][1]
                                alias = self._getTableAlias(srcIndex)
                                colName = self._getAttrName(srcIndex,colIndex)
                                sqlWhere += aliasBase+"."+colNameBase+"="+alias+"."+colName+" AND "
                                cond += 1
                    else:
                        if isinstance(t,FunctionalTerm):
                            print("WARNING: functional terms are handled as standard strings.")
                        for coord in self._bodyTermPos[t]:
                            srcIndex = coord[0]
                            colIndex = coord[1]
                            alias = self._getTableAlias(srcIndex)
                            colName = self._getAttrName(srcIndex,colIndex)
                            sqlWhere += alias + "." + colName + "='" + t.__str__().replace("\"","") +"' AND " #only strings are supported
                            cond += 1
            if cond > 0:
                sqlWhere += "NOT EXISTS ( SELECT * FROM "+ trgTable + " WHERE " + columnClause + notExistsClause
            else:
                sqlWhere += "NOT EXISTS ( SELECT * FROM "+ trgTable + " WHERE " + columnClause + notExistsClause
        sql = sqlInsert + " (" + sqlSelect + sqlFrom + sqlWhere +")"
        return sql

    def _getRecursiveRuleInSQL(self):
        trgTable = self._getTableName(self._targetPredicate)
        constCounter = 0
        sqlInsert = "INSERT INTO " + trgTable
        sqlSelect = "SELECT DISTINCT "
        notExistsClause = ""
        columnClause = ""
        for i in range(len(self._termsToSelect)):
            t = self._termsToSelect[i]
            if isinstance(t,Variable):
                assert t in self._varsFrom
                coords = self._varsFrom[t]
                srcIndex = coords[0]
                colIndex = coords[1]
                alias = self._getTableAlias(srcIndex)
                colName = self._getAttrName(srcIndex,colIndex)
                sqlSelect += alias+"."+colName+","
                columnClause += alias + "." + colName + "=" + trgTable + "." + self._targetColumns[i] + " AND "
            else:
                if isinstance(t,FunctionalTerm):
                    print("WARNING: functional terms are handled as standard strings.")
                sqlSelect += "'"+t.__str__().replace("\"","")+"' AS const"+str(constCounter)+"," #only strings are supported
                notExistsClause += trgTable+"."+self._targetColumns[i] + " = '" + t.__str__().replace("\"","") + "' AND "
                constCounter += 1
        notExistsClause += "(" + trgTable + "." + self._srcManager.getTimeStampColumnName() + " = " + self._getTimeStamp() + " OR ";
        notExistsClause += trgTable + "." + self._srcManager.getTimeStampColumnName() + " = "+self._originTimeStamp+"))";

        sqlSelect += self._getTimeStamp() + " " + " AS const" + str(constCounter) + ","
        constCounter += 1
        sqlSelect += " '"+self._newPlaceHolder+"' AS const"+str(constCounter)+","
        sqlSelect = sqlSelect[:-1]
        sqlFrom = ""
        whereClause = ""
        if len(self._bodyPreds) > 0:
            sqlFrom += " FROM "
            for i in range(len(self._bodyPreds)):
                srcTable = self._getTableName(self._bodyPreds[i])
                alias = self._getTableAlias(i)
                sqlFrom += srcTable + " AS " + alias +","
                whereClause += "(" + alias + "." + self._srcManager.getTimeStampColumnName() + " = " + self._getTimeStamp() + " OR ";
                whereClause += alias + "." + self._srcManager.getTimeStampColumnName() + " = "+self._originTimeStamp+") AND ";
            sqlFrom = sqlFrom[:-1]
        sqlWhereRecursionHandling = []
        #consider every possible combination of delta, old and new data
        for i in range(int(math.pow(2,self._nRecursiveBodyPreds)-1)):
            clause = ""
            #generate the binary representation of i to get the current combination 
            comb = bin(i+1)[2:].zfill(self._nRecursiveBodyPreds)
            x = 0
            for j in range(len(self._bodyPreds)):
                if self._recursiveBodyPreds[j] == True:
                    alias = self._getTableAlias(j)
                    if comb[x] == '0': #old
                        clause += alias+"."+self._srcManager.getRecursionLevelColumnName()+"= '"+self._oldPlaceHolder+"' AND "
                    else:
                        clause += alias+"."+self._srcManager.getRecursionLevelColumnName()+"= '"+self._deltaPlaceHolder+"' AND "
                    x += 1
            if len(clause) > 0:
                clause = clause[:-5]
                sqlWhereRecursionHandling.append(clause)
        sqlWhere = " WHERE " + whereClause
        cond = 0
        for t in self._bodyTermPos:
            if isinstance(t,Variable):
                coords = self._bodyTermPos[t]
                if len(coords) > 1:
                    srcIndexBase = coords[0][0]
                    colIndexBase = coords[0][1]
                    aliasBase = self._getTableAlias(srcIndexBase)
                    colNameBase = self._getAttrName(srcIndexBase,colIndexBase)
                    for i in range(len(coords[1:])):
                        srcIndex = coords[i+1][0]
                        colIndex = coords[i+1][1]
                        alias = self._getTableAlias(srcIndex)
                        colName = self._getAttrName(srcIndex,colIndex)
                        sqlWhere += aliasBase+"."+colNameBase+"="+alias+"."+colName+" AND "
                        cond += 1 
            else:
                if isinstance(t,FunctionalTerm):
                    print("WARNING: functional terms are handled as standard strings.")
                for coord in self._bodyTermPos[t]:
                    srcIndex = coord[0]
                    colIndex = coord[1]
                    alias = self._getTableAlias(srcIndex)
                    colName = self._getAttrName(srcIndex,colIndex)
                    sqlWhere += alias + "." + colName + "='" + t.__str__().replace("\"","") +"' AND " #only strings are supported
                    cond += 1
        sql = sqlInsert + " ("
        if len(sqlWhereRecursionHandling) == 0:
            '''if cond > 0:
                sqlWhere = sqlWhere[:-5]
            else:
                sqlWhere = ""'''
            sqlWhere += "NOT EXISTS ( SELECT * FROM " + trgTable + " WHERE " + columnClause
            sql += sqlSelect + sqlFrom + sqlWhere + notExistsClause+ ")))"
        else:
            #ERRORE....
            for clause in sqlWhereRecursionHandling:
                where = sqlWhere + clause + " AND NOT EXISTS ( SELECT * FROM " + trgTable + " WHERE " + columnClause
                sql += " (" + sqlSelect + sqlFrom + where + notExistsClause+") UNION "
            sql = sql[:-7]
            sql += ")"
        return sql

    def getNumberRowsInserted(self, cursor, parameter2=None):
        return cursor.rowcount
